
#pragma once

#include <numeric>
#include <limits>
#include <array>
#include <vector>

#include <cstdint>
#include <cassert>

namespace bitsliced
{

using bit_t = uint64_t;

static constexpr size_t width = std::numeric_limits<bit_t>::digits;
static constexpr bit_t one = ~ (bit_t) 0;

template <size_t ROWS, size_t COLS>
struct matrix_t
{
    static constexpr size_t num_rows = ROWS;
    static constexpr size_t num_cols = COLS;

    using row_t = std::array<bit_t, num_cols>;
    using col_t = std::array<bit_t, num_rows>;

    using rows_t = std::array<row_t, num_rows>;  // main matrix
    using left_t = std::array<col_t, num_rows>;  // row transform

    static inline void rref_impl(rows_t &M, left_t *T, size_t kstart, size_t kend)
    {
        static_assert(num_rows >= num_cols, "not implemented");

        kend = kend > num_cols ? num_cols : kend;

        for (size_t k = kstart; k < kend; ++k) {

#ifndef NDEBUG
for (size_t i = k+1; i < num_rows; ++i)
    for (size_t j = 0; j < k; ++j)
        assert(!M[i][j]);
#endif

            for (size_t i = 0; i < num_rows; ++i) {
                if (i == k)
                    continue;
                bit_t m = M[k][k];
                if (i < k)
                    m |= M[i][i];
                m = ~m;
                m &= M[i][k];
#if SHORTCUTS
                if (!m)
                    continue;
#endif
                for (size_t j = k; j < kend; ++j)
                    M[k][j] ^= m & M[i][j];
                if (T)
                    for (size_t j = 0; j < num_cols; ++j)
                        (*T)[k][j] ^= m & (*T)[i][j];
            }

            for (size_t i = 0; i < num_rows; ++i) {
                if (i == k)
                    continue;
                bit_t m = M[i][k];
#if SHORTCUTS
                if (!m)
                    continue;
#endif
                for (size_t j = k; j < kend; ++j)
                    M[i][j] ^= m & M[k][j];
                if (T)
                    for (size_t j = 0; j < num_cols; ++j)
                        (*T)[i][j] ^= m & (*T)[k][j];
            }

        }
    }

    static void rref(rows_t &M, size_t kstart = 0, size_t kend = std::numeric_limits<size_t>::max())
    {
        rref_impl(M, nullptr, kstart, kend);
    }

    static void rref2(rows_t &M, left_t &T, size_t kstart = 0, size_t kend = std::numeric_limits<size_t>::max())
    {
        rref_impl(M, &T, kstart, kend);
    }

    // input and output must not be the same
    static void tvec(col_t &out, col_t const &in, left_t const &T)
    {
        out = col_t {};
        for (size_t i = 0; i < num_cols; ++i)
            for (size_t j = 0; j < num_cols; ++j)
                out[i] ^= T[i][j] & in[j];
    }
};  // struct matrix_t

template <size_t LEN, size_t MAXHULLDIM, size_t HULLENUMPREC>
struct code_t
{
    static constexpr size_t length = LEN;

    using MAT = matrix_t<LEN, LEN>;

    using rows_t = MAT::rows_t;
    using left_t = MAT::left_t;
    using row_t = MAT::row_t;
    using col_t = MAT::col_t;

    static constexpr size_t max_hull_dim = MAXHULLDIM;

    struct hullrow_t {
        bit_t present;
        std::array<bit_t, LEN> row;
    };
    using hull_t = std::array<hullrow_t, max_hull_dim>;
    struct hulls_t {
        hull_t hull;
        std::array<hull_t, LEN> punctured;
    };

    // compute hull and hulls of all singly punctured codes
    // input: generator matrix in rref()
    static bit_t hull_and_punctured_hulls(hulls_t &hulls, rows_t const &code, bool hull_only = false)
    {
        // allocate these on the heap to prevent stack overflows
        auto mat = std::make_unique<rows_t>(code);
        auto left = std::make_unique<left_t>();

        // insert basis of dual in empty rows
        for (size_t i = 0; i < LEN; ++i) {
            bit_t m = ~code[i][i];
            (*mat)[i][i] ^= (*left)[i][i] = m;
            for (size_t j = 0; j < i; ++j)
                (*mat)[i][j] = (*left)[i][j] = m & code[j][i];
        }

        // row-reduce again
        MAT::rref2(*mat, *left);

        hulls = hulls_t {};

        // compute hull
        bit_t is_okay = one;
        {
            hull_t &hull = hulls.hull;

            std::array<bit_t, max_hull_dim+1> idx {one};
            for (size_t i = 0; i < LEN; ++i) {
                bit_t m = ~(*mat)[i][i];
                for (ssize_t k = max_hull_dim-1; k >= 0; --k) {
                    bit_t mm = m & idx[k];
                    for (size_t j = 0; j < i; ++j)
                        hull[k].row[j] ^= mm & (*mat)[j][i];
                    hull[k].row[i] ^= mm;
                    idx[k+0] ^= mm;
                    idx[k+1] ^= mm;
                }
            }
            is_okay &= ~idx[max_hull_dim];

            for (size_t i = 0; i < max_hull_dim; ++i)
                hull[i].present = idx[i+1];
            for (ssize_t i = max_hull_dim-2; i >= 0; --i)
                hull[i].present |= hull[i+1].present;

#ifndef NDEBUG
for (size_t i = 0; i < max_hull_dim; ++i) {
    bit_t any = 0;
    for (size_t j = 0; j < LEN; ++j) {
        any |= hull[i].row[j];
    }
    assert((~hull[i].present | any) == one);
}
#endif
        }

        if (hull_only)
            return is_okay;

        // compute punctured hulls
        // see Sendrier, The Support Splitting Algorithm, §5.3.1
        for (size_t loc = 0; loc < LEN; ++loc) {

            hull_t &puncthull = hulls.punctured[loc] = hulls.hull;
            bit_t done {0};

            // in support
            for (ssize_t k = max_hull_dim - 1; k >= 0; --k) {
#if SHORTCUTS
                if (done == one)
                    break;
#endif
                bit_t m = ~done & puncthull[k].present & puncthull[k].row[loc];
                for (size_t i = 0; i < (size_t) k; ++i) {
                    bit_t mm = m & puncthull[i].row[loc];
                    for (size_t j = 0; j < LEN; ++j)
                        puncthull[i].row[j] ^= mm & puncthull[k].row[j];
                }
                puncthull[k].present &= ~m;
                done |= m;
            }

            // not in support
            bit_t m = ~(*left)[loc][loc];

            for (size_t i = 0; i < max_hull_dim; ++i) {
                m &= ~done;
#if SHORTCUTS
                if (!m)
                    break;
#endif
                bit_t mm = m & ~puncthull[i].present;
                for (size_t j = 0; j < LEN; ++j) {
                    puncthull[i].row[j] ^= mm & (*left)[loc][j];
                    puncthull[i].present |= mm & puncthull[i].row[j];
                }
                done |= mm & puncthull[i].present;
            }

#ifndef NDEBUG
for (size_t i = 0; i < max_hull_dim; ++i) {
    bit_t any = 0;
    for (size_t j = 0; j < LEN; ++j) {
        any |= puncthull[i].row[j];
    }
    assert((~puncthull[i].present | any) == one);
}
{
    rows_t punctcode = code;
    for (size_t i = 0; i < LEN; ++i)
        punctcode[i][loc] = 0;
    MAT::rref(punctcode);
    hulls_t mypuncthull;
    hull_and_punctured_hulls(mypuncthull, punctcode, true);
    rows_t x {}, y {};
    for (size_t r = 0, i = 0; i < max_hull_dim; ++i) {
        if (hulls.punctured[loc][i].present & 1) {
            for (size_t j = 0; j < LEN; ++j)
                x[r][j] = hulls.punctured[loc][i].row[j] & 1;
            ++r;
        }
    }
    for (size_t r = 0, i = 0; i < max_hull_dim; ++i) {
        if (mypuncthull.hull[i].present & 1) {
            for (size_t j = 0; j < LEN; ++j)
                y[r][j] = mypuncthull.hull[i].row[j] & 1;
            ++r;
        }
    }
    MAT::rref(x);
    MAT::rref(y);
    assert(!(is_okay & 1) || x == y);
}
#endif
        }

        return is_okay;
    }

    template <size_t SZ>
    struct num_t {
        std::array<bit_t, SZ> bits {};
        bit_t overflow {0};
        static constexpr size_t max = ((size_t) 1 << SZ) - 1;

        void inc(bit_t carry)
        {
            for (size_t i = 0; i < SZ; ++i) {
                bit_t newcarry = bits[i] & carry;
                bits[i] ^= carry;
                carry = newcarry;
            }
            overflow |= carry;
        }

        bit_t eq(uint64_t other) const
        {
            static_assert(SZ < 64);
            assert(other <= max);
            bit_t neq = overflow;
            for (size_t i = 0; i < SZ; ++i) {
                neq |= bits[i] ^ ((other >> i) & 1 ? one : bit_t{0});
            }
            return ~neq;
        }
    };

    static constexpr size_t hull_enum_prec = HULLENUMPREC;

    struct hullenum_t {
        using count_t = num_t<hull_enum_prec>;
        std::array<count_t, LEN+1> counts;

        template <class I>
        bit_t eq(std::array<I, LEN+1> const &other) const
        {
            bit_t iseq = one;
            for (size_t i = 0; i <= LEN; ++i)
                iseq &= counts[i].eq(other[i]);
            return iseq;
        }
    };
    struct hullenums_t {
        hullenum_t hullenum;
        std::array<hullenum_t, LEN> punctured;
    };

    // compute enumerator polynomial from hull using Gray code
    static void enumerator(hullenum_t &res, hull_t const &hull)
    {
        res = hullenum_t {};
        res.counts[0].inc(one);

        std::array<bool, max_hull_dim> gray {};
        row_t row {};

        for (size_t step = 1; step < ((size_t) 1 << max_hull_dim); ++step) {

            size_t const idx = std::countr_zero(step);
            gray[idx] = !gray[idx];

#if SHORTCUTS
            if (hull[idx].present)
                for (size_t j = 0; j < LEN; ++j)
                    row[j] ^= hull[idx].row[j];
#else
            for (size_t j = 0; j < LEN; ++j)
                row[j] ^= hull[idx].present & hull[idx].row[j];
#endif

            bit_t m = one;
            for (size_t i = 0; i < max_hull_dim; ++i)
                if (gray[i])
                    m &= hull[i].present;
#if SHORTCUTS
            if (!m)
                continue;
#endif

            num_t<64-std::countl_zero<uint64_t>(LEN)> ent;
            for (size_t j = 0; j < LEN; ++j)
                ent.inc(row[j]);
            for (size_t w = 0; w <= LEN; ++w)
                res.counts[w].inc(m & ent.eq(w));
        }
    }

    static void enumerators(hullenums_t &enums, hulls_t const &hulls)
    {
        enumerator(enums.hullenum, hulls.hull);
        for (size_t i = 0; i < LEN; ++i)
            enumerator(enums.punctured[i], hulls.punctured[i]);
    }

    static bit_t check_filter(rows_t const &mat, std::array<uint8_t,LEN+1> const &the_hullenum, std::vector<std::pair<bool,std::array<uint8_t,LEN+1>>> const &fast_filter)
    {
        hulls_t hulls;
        auto is_okay = hull_and_punctured_hulls(hulls, mat);

        hullenums_t enums;

        enumerator(enums.hullenum, hulls.hull);
        is_okay &= enums.hullenum.eq(the_hullenum);

        for (size_t i = 0; i < LEN; ++i) {
#if SHORTCUTS
            if (!is_okay)
                break;
#endif

            enumerator(enums.punctured[i], hulls.punctured[i]);

            for (auto const &[b,v]: fast_filter) {
                if (b)
                    continue;
                is_okay &= ~enums.punctured[i].eq(v);
            }
        }

        for (auto const &[b,v]: fast_filter) {
            if (!b)
                continue;
            bitsliced::bit_t anyeq {0};
            for (auto const &henum: enums.punctured)
                anyeq |= henum.eq(v);
            is_okay &= anyeq;
        }

        return is_okay;
    }
};  // struct code_t

};  // namespace bitsliced

