
#include "goppa.hpp"

#include <random>
#include <sstream>

#include <cassert>

////////////////////////////////////////////////////////////////

#include <NTL/GF2XFactoring.h>
#include <NTL/GF2EX.h>
#include <NTL/GF2EXFactoring.h>
#include <NTL/matrix.h>
#include <NTL/mat_GF2E.h>

void init_ntl_random()
{
    auto const urand = []() { return std::random_device{}(); };
    std::seed_seq s {urand(), urand(), urand(), urand()};
    std::mt19937 rng(s);
    NTL::ZZ seed;
    for (size_t i = 0; i < 4; ++i)
        seed = seed << 32 ^ rng();
    NTL::SetSeed(seed);
}

static NTL::GF2E &fgen()
{
    thread_local bool init = false;
    thread_local NTL::GF2E a;
    if (init)
        return a;
    NTL::GF2X x;
    NTL::SetX(x);
    NTL::conv(a, x);
    init = true;
#ifndef NDEBUG
    {
        // check assumption: x is a primitive element
        NTL::GF2E tmp = a;
        assert(!NTL::IsZero(tmp));
        for (size_t i = 1; i < NTL::GF2E::cardinality()-1; ++i) {
            assert(!NTL::IsOne(tmp));
            tmp *= a;
        }
    }
#endif
    return a;
}

struct goppa_generator::data_t {
    size_t m, t;
    NTL::GF2EX g;
    std::vector<NTL::GF2E> felts;
};

goppa_generator::goppa_generator(std::vector<bool> const &fieldpoly, size_t t)
{
    NTL::GF2X ntl_poly;

    data = std::make_unique<data_t>();

    assert(!fieldpoly.empty() && fieldpoly.back());
    for (size_t i = 0; i < fieldpoly.size(); ++i)
        NTL::SetCoeff(ntl_poly, i, fieldpoly[i]);
    assert(NTL::IterIrredTest(ntl_poly));

    NTL::GF2E::init(ntl_poly);

    data->m = NTL::deg(ntl_poly);
    data->t = t;
    assert(data->t >= 1);

    data->g = NTL::BuildIrred_GF2EX(data->t);

    data->felts.resize(1);
    for (size_t i = 0; i < NTL::GF2E::cardinality()-1; ++i)
        data->felts.push_back(NTL::power(fgen(), i));
}

goppa_generator::~goppa_generator()
{
    data.reset();
}

#include <algorithm>

// the Goppa points yielding the given columns do not need to be passed
// explicitly: they are always just [0, 1, a, a^2, a^3, ..., a^(2^m-2)]
// where a is the generator (i.e., image of x) of our model of GF(2^m)
void goppa_generator::next(std::string &poly, std::vector<std::vector<bool>> &cols)
{
    assert(NTL::deg(data->g) == (ssize_t) data->t);
    assert(NTL::GF2E::degree() == (ssize_t) data->m);

    {
        std::ostringstream ss;
        ss << "[";
        for (size_t i = 0; i <= data->t; ++i) {
            size_t which = std::find(data->felts.begin(), data->felts.end(), data->g[i]) - data->felts.begin();
            ss << (i ? "," : "") << fmt_felt(which);
        }
        ss << "]";
        poly = ss.str();
    }

    size_t q;
    NTL::conv(q, NTL::GF2E::cardinality());

    cols.resize(q);
    for (auto &col: cols)
        col.resize(data->t * data->m);

    for (size_t k = 0; k < q; ++k) {
        auto cur = NTL::inv(NTL::eval(data->g, data->felts[k]));
        for (size_t i = 0; i < data->t; ++i) {
            if (i)
                NTL::mul(cur, cur, data->felts[k]);
            for (size_t j = 0; j < data->m; ++j)
                cols[k][data->m*i+j] = NTL::IsOne(NTL::coeff(NTL::rep(cur), j));
        }
    }

    data->g = NTL::BuildRandomIrred(data->g);
}

std::string goppa_generator::fmt_felt(size_t which) const
{
    switch (which) {
        case 0:
            assert(NTL::IsZero(data->felts[which]));
            return "0";
        case 1:
            assert(NTL::IsOne(data->felts[which]));
            return "1";
        case 2:
            assert(data->felts[which] == fgen());
            return "a";
        default:
            assert(data->felts[which] == NTL::power(fgen(), which-1));
            return "a^" + std::to_string(which-1);
    };
}

