
#include <cassert>
#include <cstdint>

#include <iostream>
#include <iomanip>

////////////////////////////////////////////////////////////////

[[maybe_unused]]
static inline uint64_t rdtsc(void)
{
    uint32_t hi, lo;
    __asm__ __volatile__ ("rdtsc" : "=a"(lo), "=d"(hi));
    return lo | (uint64_t) hi << 32;
}

////////////////////////////////////////////////////////////////

#include <atomic>
#include <thread>

std::atomic<bool> terminate = false;

const unsigned num_threads = std::thread::hardware_concurrency() ? std::thread::hardware_concurrency() : 1;

#include <csignal>

void handler(int num)
{
    (void) num;
    terminate = true;
}

std::mutex mtx;

std::atomic<size_t> running = 0;

std::atomic<size_t> the_count = 0;
std::atomic<size_t> the_good_count = 0;

////////////////////////////////////////////////////////////////

#include <cmath>
#include <algorithm>

static std::string thsep(uint64_t v)
{
    if (!v)
        return "0";
    std::ostringstream ss;
    for (size_t i = 0; v; ++i) {
        ss << v % 10;
        (v /= 10) && i % 3 == 2 && ss << ',';
    }
    auto s = ss.str();
    std::reverse(s.begin(), s.end());
    return s;
}

static std::string fmtsec(double seconds)
{
    unsigned long dcount = std::lround(seconds) / 86400;
    unsigned long hcount = std::lround(seconds) % 86400 / 3600;
    unsigned long mcount = std::lround(seconds) % 3600 / 60;
    unsigned long scount = std::lround(seconds) % 60;
    return (std::ostringstream{} << dcount << "d " << hcount << "h " << mcount << "m " << scount << "s").str();
}

void status_thread(double const expected_work)
{
    ++running;

    std::chrono::time_point<std::chrono::steady_clock> const start_time = std::chrono::steady_clock::now();

    while (!terminate) {
        std::this_thread::sleep_for(std::chrono::milliseconds(500));

        auto delta = std::chrono::steady_clock::now() - start_time;
        double elapsed = 1e-9 + std::chrono::duration_cast<std::chrono::duration<double>>(delta).count();
        double expected = expected_work / (the_count / elapsed);

        std::ostringstream ss;
        ss << "\x1b[K  "
            << "\x1b[31m" << fmtsec(expected) << "\x1b[0m | "
            << "\x1b[33m" << fmtsec(elapsed) << "\x1b[0m | "
            << "\x1b[36m" << thsep(the_count) << "\x1b[0m "
            << "\x1b[1;36m" << thsep(std::lround(the_count/elapsed)) << "/s\x1b[0m | "
            << "\x1b[34m" << thsep(the_good_count) << "\x1b[0m "
            << "\x1b[1;34m" << thsep(std::lround(the_good_count/elapsed)) << "/s\x1b[0m";

        std::lock_guard lock(mtx);
        std::cerr << ss.str() << "\r" << std::flush;
    }

    --running;
}

////////////////////////////////////////////////////////////////

#define SHORTCUTS 1

#include "bitsliced.hpp"
#include "generated.hpp"

#include <functional>
#include <random>

template <class CODE>
using callback_t = std::function<void(typename CODE::rows_t const &, std::array<size_t, CODE::length>)>;

template <class CODE>
void brute_columns(std::vector<typename CODE::col_t> const &sliced_cols, callback_t<CODE> const &fun)
{
    typename CODE::rows_t mat {};
    std::array<size_t, CODE::length> idxs;
    size_t count = 0;

    typename CODE::left_t left {};
    for (size_t j = 0; j < CODE::length; ++j)
        left[j][j] = bitsliced::one;

    using rec_t = std::function<void(size_t)>;
    rec_t const rec = [&](size_t depth) -> void
    {
        if (terminate || count > 9'000)  //XXX arbitrary
            return;

        assert(depth <= CODE::length);
        if (depth == CODE::length) {
            // base case
            ++count;
#ifndef NDEBUG
{
auto mat0 = std::make_unique<typename CODE::rows_t>();
for (size_t i = 0; i < CODE::length; ++i)
    for (size_t j = 0; j < CODE::length; ++j)
        (*mat0)[i][j] = sliced_cols[idxs[j]][i];
for (size_t j = 0; j < CODE::length; ++j) {
    typename CODE::col_t col, col0;
    for (size_t i = 0; i < CODE::length; ++i)
        col0[i] = (*mat0)[i][j];
    CODE::MAT::tvec(col, col0, left);
    for (size_t i = 0; i < CODE::length; ++i)
        assert(col[i] == mat[i][j]);
}
auto mymat = std::make_unique<typename CODE::rows_t>(*mat0);
CODE::MAT::rref(*mymat);
assert(*mymat == mat);
}
#endif

            fun(mat, idxs);

            return;
        }

        assert(!depth || idxs[depth-1] + (CODE::length-depth) < sliced_cols.size());

        auto &idx = idxs[depth];
        for (idx = depth ? idxs[depth-1]+1 : 0; idx + (CODE::length-depth) <= sliced_cols.size(); ++idx) {
            typename CODE::col_t col;
            CODE::MAT::tvec(col, sliced_cols[idx], left);
            for (size_t i = 0; i < CODE::length; ++i)
                mat[i][depth] = col[i];
            CODE::MAT::rref2(mat, left, depth, depth+1);
            rec(depth+1);
        }
    };

    rec(0);
}

////////////////////////////////////////////////////////////////

#include "goppa.hpp"
#include "supportsplitting.hpp"

void worker()
{
    ++running;

    init_ntl_random();

    auto const urand = []() { return std::random_device{}(); };
    std::seed_seq s {urand(), urand(), urand(), urand()};
    std::mt19937 rng(s);

    goppa_generator gg(POLY, T);

    using CODE = bitsliced::code_t<N, MAX_HULL_DIM, HULL_ENUM_PREC>;

    std::array<std::string, bitsliced::width> polys;
    std::array<std::vector<std::vector<bool>>, bitsliced::width> colss;
    while (!terminate) {
        for (size_t k = 0; k < bitsliced::width; ++k)
            gg.next(polys[k], colss[k]);

        // permute
        std::vector<size_t> perm((size_t) 1 << M);
        std::iota(perm.begin(), perm.end(), 0);
        std::shuffle(perm.begin(), perm.end(), rng);

        // slice
        std::vector<CODE::col_t> sliced_cols((size_t) 1 << M);
        for (size_t i = 0; i < sliced_cols.size(); ++i)
            for (size_t k = 0; k < bitsliced::width; ++k)
                for (size_t j = 0; j < T*M; ++j)
                    sliced_cols[i][j] |= (bitsliced::bit_t) colss[k][perm[i]][j] << k;

        callback_t<CODE> const fun = [&](CODE::rows_t const &mat, std::array<size_t, CODE::length> const &idxs)
        {
            bitsliced::bit_t is_okay = CODE::check_filter(mat, the_hullenum, fast_filter);

            the_count += bitsliced::width;
            the_good_count += std::popcount(is_okay);

            for (size_t k = 0; k < bitsliced::width; ++k) {

                if (!((is_okay >> k) & 1))
                    continue;

                std::array<std::array<bool,CODE::length>,CODE::length> code;
                for (size_t i = 0; i < CODE::length; ++i)
                    for (size_t j = 0; j < CODE::length; ++j)
                        code[i][j] = (sliced_cols[idxs[j]][i] >> k) & 1;

                if (auto equiv = supportsplitting<N>::run(code, supportsplitting_data)) {

                    if (terminate.exchange(true))
                        break;

                    std::vector<size_t> altogether(N);  // final permutation
                    for (size_t i = 0; i < N; ++i)
                        altogether[i] = perm[idxs[equiv->at(i)]];

                    while (running > 1)
                        std::this_thread::sleep_for(std::chrono::milliseconds(1));

                    std::ostringstream ss;
                    ss << polys[k] << "\n[";
                    for (size_t i = 0; i < CODE::length; ++i)
                        ss << (i ? "," : "") << gg.fmt_felt(altogether[i]);
                    ss << "]";
                    std::cerr << "\n\x1b[32m" << ss.str() << "\x1b[0m" << std::endl;
                    std::cout << ss.str() << std::endl;
                    break;
                }
            }
        };

        brute_columns<CODE>(sliced_cols, fun);
    }

    --running;
}

////////////////////////////////////////////////////////////////

int main()
{
    signal(SIGINT, &handler);

    {
        std::vector<std::jthread> threads;
        while (threads.size() < num_threads)
            threads.emplace_back(worker);
        threads.emplace_back(status_thread, WORK);
    }
}

