
#include <cassert>
#include <cstdint>

#include <iostream>
#include <iomanip>

////////////////////////////////////////////////////////////////

[[maybe_unused]]
static inline uint64_t rdtsc(void)
{
    uint32_t hi, lo;
    __asm__ __volatile__ ("rdtsc" : "=a"(lo), "=d"(hi));
    return lo | (uint64_t) hi << 32;
}

////////////////////////////////////////////////////////////////

#include <atomic>
#include <thread>
#include <execution>
#include <future>

const unsigned num_threads = std::thread::hardware_concurrency() ? std::thread::hardware_concurrency() : 1;

////////////////////////////////////////////////////////////////

struct {
    size_t howmany = 100000;
    size_t m, t, n;
    std::vector<bool> poly;
    std::vector<std::vector<bool>> the_code;
} params;

void read_input()
{
    std::vector<std::string> lines;
    for (std::string line; std::getline(std::cin, line); )
        lines.push_back(std::move(line));

    for (char c: lines.back()) {
        if (c == '0')
            params.poly.push_back(0);
        if (c == '1')
            params.poly.push_back(1);
    }
    lines.pop_back();

    for (auto const &l: lines) {
        params.the_code.emplace_back();
        for (char c: l) {
            if (c == '0')
                params.the_code.back().push_back(0);
            if (c == '1')
                params.the_code.back().push_back(1);
        }
    }

    params.m = params.poly.size() - 1;
    assert(params.m >= 2);
    params.t = params.the_code.size() / params.m;
    assert(params.t >= 1);
    params.n = params.the_code[0].size();
    assert(params.n >= 1);
    assert(params.the_code.size() == params.t * params.m);
#ifndef NDEBUG
    for (auto const &row: params.the_code)
        assert(row.size() == params.n);
#endif
}

////////////////////////////////////////////////////////////////

#include "goppa.hpp"
#include "bitsliced.hpp"
#include "supportsplitting.hpp"
#include "hash.hpp"

#include <random>
#include <map>

template <size_t N>
using hashes_map_t = std::map<uint64_t, single_hullenum_t<N>>;
template <size_t N>
using enums_list_t = std::vector<std::pair<uint64_t, std::array<uint64_t,N>>>;

template <size_t N>
void worker(std::mutex &mtx, std::atomic<size_t> &count, hashes_map_t<N> &hashes, enums_list_t<N> &enums)
{
    using CODE = bitsliced::code_t<N, 8, 8>;

    init_ntl_random();

    auto const urand = []() { return std::random_device{}(); };
    std::seed_seq s {urand(), urand(), urand(), urand()};
    std::mt19937 rng(s);

    hashes_map_t<N> myhashes;
    enums_list_t<N> myenums;

    goppa_generator gg(params.poly, params.t);

    std::array<std::vector<std::vector<bool>>, bitsliced::width> colss;
    while (true) {
        for (auto &cols: colss) {
            std::string tmp;
            gg.next(tmp, cols);
        }

        // slice
        std::vector<typename CODE::col_t> sliced_cols((size_t) 1 << params.m);
        for (size_t i = 0; i < sliced_cols.size(); ++i)
            for (size_t k = 0; k < bitsliced::width; ++k)
                for (size_t j = 0; j < params.the_code.size(); ++j)
                    sliced_cols[i][j] |= (bitsliced::bit_t) colss[k][i][j] << k;

        // shuffle
        std::shuffle(sliced_cols.begin(), sliced_cols.end(), rng);

        // build code
        typename CODE::rows_t sliced_mat;
        for (size_t i = 0; i < N; ++i)
            for (size_t j = 0; j < N; ++j)
                sliced_mat[i][j] = sliced_cols[j][i];
        CODE::MAT::rref(sliced_mat);

        // compute hull enumerators
        typename CODE::hulls_t sliced_hulls;
        auto is_okay = CODE::hull_and_punctured_hulls(sliced_hulls, sliced_mat);
        typename CODE::hullenums_t sliced_enums;
        CODE::enumerators(sliced_enums, sliced_hulls);

        // unslice
        for (size_t j = 0; j <= N; ++j) {
            is_okay &= ~sliced_enums.hullenum.counts[j].overflow;
            for (auto const &t: sliced_enums.punctured)
                is_okay &= ~t.counts[j].overflow;
        }

        std::vector<std::pair<single_hullenum_t<N>, std::array<single_hullenum_t<N>,N>>> data;
        data.reserve(bitsliced::width);

        for (size_t k = 0; k < bitsliced::width; ++k) {
            auto m = (bitsliced::bit_t) 1 << k;

            if (!(is_okay & m))
                continue;

            auto &[u,vs] = data.emplace_back();

            for (size_t j = 0; j <= N; ++j) {
                auto const conv = [&](auto const &o) {
                    size_t r = 0;
                    for (size_t t = 0; t < o.bits.size(); ++t)
                        r |= ((o.bits[t] >> k) & 1) << t;
                    return r;
                };
                u[j] = conv(sliced_enums.hullenum.counts[j]);
                for (size_t i = 0; i < vs.size(); ++i)
                    vs[i][j] = conv(sliced_enums.punctured[i].counts[j]);
            }
        }

        auto const inshash = [&myhashes](single_hullenum_t<N> const &v) -> uint64_t {
            auto h = myhash(v);
            if (auto [it,_] = myhashes.insert({h, v}); it->second != v)
                throw std::logic_error("hash collision");
            return h;
        };

        for (auto const &[u,vs]: data) {
            typename enums_list_t<N>::value_type r;
            r.first = inshash(u);
            assert(vs.size() == r.second.size());
            for (size_t i = 0; i < vs.size(); ++i)
                r.second[i] = inshash(vs[i]);
            myenums.push_back(std::move(r));
        }
        if ((count += data.size()) >= params.howmany)
            break;
    }

    std::lock_guard lock(mtx);

    for (auto const &tup: myhashes)
        if (auto [it,_] = hashes.insert(tup); it->second != tup.second)
            throw std::logic_error("hash collision");

    std::copy(std::make_move_iterator(myenums.begin()), std::make_move_iterator(myenums.end()), std::back_inserter(enums));
}

////////////////////////////////////////////////////////////////

#include <set>

template <size_t N>
struct filter_data_t {
    size_t max_hull_dim, hull_enum_prec;
    single_hullenum_t<N> hullenum;
    std::vector<std::pair<bool, single_hullenum_t<N>>> filter;
};

template <size_t N>
filter_data_t<N> precompute_filter(std::array<std::array<bool,N>,N> const &code)
{
    auto [hullenum, puncturedenums] = compute_hullenum_and_puncturedenums<N>(code);

    size_t max_hull_card = std::accumulate(hullenum.begin(), hullenum.end(), (size_t) 0);
    for (auto const &e: puncturedenums) {
        auto card = std::accumulate(e.begin(), e.end(), (size_t) 0);
        max_hull_card = std::max(max_hull_card, card);
    }
    size_t max_hull_dim = 64 - std::countl_zero<uint64_t>(max_hull_card);

    size_t hull_enum_bound = *std::max_element(hullenum.begin(), hullenum.end());
    for (auto const &e: puncturedenums) {
        size_t bound = *std::max_element(e.begin(), e.end());
        hull_enum_bound = std::max(hull_enum_bound, bound);
    }
    size_t hull_enum_prec = 64 - std::countl_zero<uint64_t>(hull_enum_bound);
    assert(hull_enum_bound < ((size_t) 1 << hull_enum_prec));

    hashes_map_t<N> hashes;
    enums_list_t<N> enums;
    enums.reserve(params.howmany);
    {
        std::mutex mtx;
        std::atomic<size_t> count = 0;
        std::vector<std::jthread> threads;

        while (threads.size() < num_threads)
            threads.emplace_back(worker<N>, std::ref(mtx), std::ref(count), std::ref(hashes), std::ref(enums));
    }

    size_t const total = enums.size();

    std::set<uint64_t> yes, no;
    for (auto const &e: puncturedenums) {
        auto h = myhash(e);
        if (auto [it,_] = hashes.insert({h, e}); it->second != e)
            throw std::logic_error("hash collision");
        yes.insert(myhash(e));
    }
    for (auto const &[_,pes]: enums)
        for (auto const &h: pes)
            if (yes.find(h) == yes.end())
                no.insert(h);

    using filter_t = std::vector<std::pair<bool, uint64_t>>;
    auto const run_filter = [&](filter_t const &filter, enums_list_t<N> const &inenums, enums_list_t<N> *outenums=nullptr) -> size_t {
#ifndef NDEBUG
{
    bool chk = true;
    for (auto const &[b,h]: filter) {
        bool any = false;
        for (auto const &e: puncturedenums)
            any |= myhash(e) == h;
        chk &= any == b;
    }
    assert(chk);
}
#endif
        size_t r = 0;
        for (auto const &[hh,phs]: inenums) {
            bool ok = myhash(hullenum) == hh;
            for (auto const &[b,h]: filter)
                if (!(ok &= (std::find(phs.begin(), phs.end(), h) != phs.end()) == b))
                    break;
            if (!ok)
                continue;
            ++r;
            if (outenums)
                outenums->push_back(std::make_pair(hh, phs));
        }
        assert(!outenums || outenums->size() == r);
        return r;
    };

    filter_t cur_filter;
    enums_list_t<N> cur_enums;
    run_filter(cur_filter, enums, &cur_enums);
std::cerr << "\x1b[33mPr ~ 1/" << (double) total / run_filter(cur_filter, cur_enums) << "\x1b[0m" << std::endl;

    std::vector<uint64_t> yesno;
    for (auto const &h: yes)
        yesno.push_back(h);
    for (auto const &h: no)
        yesno.push_back(h);

    while (!yesno.empty()) {

        std::vector<size_t> counts(yesno.size());
        std::transform(std::execution::par_unseq, yesno.begin(), yesno.end(), counts.begin(),
            [&](uint64_t h) -> size_t {
                return run_filter({std::make_pair(yes.find(h) != yes.end(), h)}, cur_enums);
            });
        uint64_t h = yesno[std::min_element(counts.begin(), counts.end()) - counts.begin()];
        bool b = yes.find(h) != yes.end();
        assert(hashes.find(h) != hashes.end());

        auto new_filter = cur_filter;
        new_filter.push_back(std::make_pair(b, h));
        if (run_filter(new_filter, cur_enums) >= run_filter(cur_filter, cur_enums))
            break;

        cur_filter = std::move(new_filter);
        enums_list_t<N> new_enums;
        run_filter(cur_filter, cur_enums, &new_enums);
        cur_enums = std::move(new_enums);

std::cerr << "\x1b[33mPr ~ 1/" << (double) total / run_filter(cur_filter, cur_enums) << "\x1b[0m" << std::endl;
    }

    std::vector<std::pair<bool, single_hullenum_t<N>>> filter;
    for (auto const &[b,h]: cur_filter)
    {
std::cerr << (b ? "\x1b[32my\x1b[0m " : "\x1b[31mn\x1b[0m ");
        auto const &e = hashes.at(h);
for (size_t i = 0; i < e.size(); ++i)
    std::cerr << (i ? "," : "") << +e[i];
std::cerr << "\n";
        filter.emplace_back(b, e);
    }

    return {max_hull_dim, hull_enum_prec, hullenum, filter};
}


////////////////////////////////////////////////////////////////

template <size_t N>
void precompute()
{
    assert(params.n == N);

    std::array<std::array<bool,N>,N> code {};
    assert(params.the_code.size() <= N);
    for (size_t i = 0; i < params.the_code.size(); ++i) {
        assert(params.the_code[i].size() == N);
        std::copy(params.the_code[i].begin(), params.the_code[i].end(), code[i].begin());
    }

    auto async_filter_precomp = std::async(std::launch::async, precompute_filter<N>, code);
    auto async_supportsplitting_precomp = std::async(std::launch::async, supportsplitting<N>::precompute, code);

    auto filter_precomp = async_filter_precomp.get();
    auto supportsplitting_precomp = async_supportsplitting_precomp.get();

#ifndef NDEBUG
assert(supportsplitting<N>::run(code, supportsplitting_precomp));
{
std::vector<size_t> perm(N);
std::iota(perm.begin(), perm.end(), 0);
for (size_t t = 0; t < 100000; ++t)
    std::swap(perm[rand()%N], perm[rand()%N]);
auto code2 = code;
for (size_t i = 0; i < N; ++i)
    for (size_t j = 0; j < N; ++j)
        code2[i][perm[j]] = code[i][j];
assert(supportsplitting<N>::run(code2, supportsplitting_precomp));
}
#endif

    std::cout << "size_t constexpr MAX_HULL_DIM = " << filter_precomp.max_hull_dim << ";\n";
    std::cout << "size_t constexpr HULL_ENUM_PREC = " << filter_precomp.hull_enum_prec << ";\n";
    std::cout << "\n";
    std::cout << "std::array<uint8_t,N+1> const the_hullenum\n";
    std::cout << "        {";
    for (size_t i = 0; i < filter_precomp.hullenum.size(); ++i)
        std::cout << (i ? "," : "") << +filter_precomp.hullenum[i];
    std::cout << "};\n";
    std::cout << "\n";
    std::cout << "std::vector<std::pair<bool,std::array<uint8_t,N+1>>> const fast_filter {\n";
    for (auto const &[b,e]: filter_precomp.filter) {
        std::cout << "    {" << +b << ", {";
        for (size_t i = 0; i < e.size(); ++i)
            std::cout << (i ? "," : "") << +e[i];
        std::cout << "}},\n";
    }
    std::cout << "};\n";
    std::cout << "\n";
    std::cout << "std::pair<std::vector<std::set<uint64_t>>, std::vector<std::map<uint64_t, size_t>>> const supportsplitting_data {\n";
    std::cout << "    {\n";
    for (auto const &ks: supportsplitting_precomp.first) {
        std::cout << "        {\n";
        for (auto k: ks)
            std::cout << "            0x" << std::hex << std::setw(16) << std::setfill('0') << k << ",\n";
        std::cout << "        },\n";
    }
    std::cout << "    }, {\n";
    for (auto const &tab: supportsplitting_precomp.second) {
        std::cout << "        {\n";
        for (auto const &[k,cnt]: tab) {
            std::cout << "            {0x" << std::hex << std::setw(16) << std::setfill('0') << k << ", " << std::dec << cnt << "},\n";
        }
        std::cout << "        },\n";
    }
    std::cout << "    },\n";
    std::cout << "};\n";
    std::cout << std::endl;
}

////////////////////////////////////////////////////////////////

template <typename I>
static I binom(size_t const &N, size_t const &K)
{
    if (K > N)
        return 0;
    if (K == 0 || K == N)
        return 1;
    std::vector<std::vector<I>> tab(N);
    for (size_t n = 0; n < N; ++n) {
        tab[n].resize(n+1);
        tab[n][0] = tab[n][n] = 1;
        for (size_t k = 1; k < n; ++k)
            tab[n][k] = tab[n-1][k-1] + tab[n-1][k];
    }
    assert(0 < K && K < N);
    return tab[N-1][K-1] + tab[N-1][K];
}

static double calculate_work()
{
    size_t const q = (size_t) 1 << params.m;
    double polys = std::pow(q, params.t) / params.t;  // approximate
    auto const supps = binom<double>(q, params.n);
    auto const equiv = q * (q-1) * params.m;
    return polys * supps / equiv;
}

int main(int argc, char **argv)
{
    if (argc > 1)
        params.howmany = strtoull(argv[1], nullptr, 0);

    read_input();

    std::cout << "\n";
    std::cout << "#pragma once\n";
    std::cout << "\n";
    std::cout << "#include <cstdint>\n";
    std::cout << "#include <vector>\n";
    std::cout << "#include <set>\n";
    std::cout << "#include <map>\n";
    std::cout << "\n";
    std::cout << "std::vector<bool> const POLY {";
    for (size_t i = 0; i < params.poly.size(); ++i)
        std::cout << (i ? ", " : "") << params.poly[i];
    std::cout << "};\n";
    std::cout << "size_t constexpr M = " << params.m << ";\n";
    std::cout << "size_t constexpr T = " << params.t << ";\n";
    std::cout << "size_t constexpr N = " << params.n << ";\n";
    std::cout << "\n";

    {
        auto work = calculate_work();
        std::ostringstream ss;
        ss << std::fixed << std::setprecision(0) << work;
        std::cout << "double constexpr WORK = " << ss.str() << ";  // 2^" << std::log2(work) << "\n";
        std::cout << "\n";
    }

    void (*fun)() = nullptr;
    switch (params.n) {
        case 21:   fun = precompute<21>;   break;
        case 23:   fun = precompute<23>;   break;
        case 26:   fun = precompute<26>;   break;
        case 27:   fun = precompute<27>;   break;
        case 28:   fun = precompute<28>;   break;
        case 32:   fun = precompute<32>;   break;
        case 53:   fun = precompute<53>;   break;
        case 54:   fun = precompute<54>;   break;
        case 55:   fun = precompute<55>;   break;
        case 56:   fun = precompute<56>;   break;
        case 57:   fun = precompute<57>;   break;
        case 58:   fun = precompute<58>;   break;
        case 59:   fun = precompute<59>;   break;
        case 60:   fun = precompute<60>;   break;
        case 61:   fun = precompute<61>;   break;
        case 62:   fun = precompute<62>;   break;
        case 63:   fun = precompute<63>;   break;
        case 73:   fun = precompute<73>;   break;
        case 79:   fun = precompute<79>;   break;
        case 83:   fun = precompute<83>;   break;
        case 88:   fun = precompute<88>;   break;
        case 94:   fun = precompute<94>;   break;
        case 96:   fun = precompute<96>;   break;
        case 97:   fun = precompute<97>;   break;
        case 98:   fun = precompute<98>;   break;
        case 100:  fun = precompute<100>;  break;
        case 103:  fun = precompute<103>;  break;
        case 105:  fun = precompute<105>;  break;
        case 106:  fun = precompute<106>;  break;
        case 108:  fun = precompute<108>;  break;
        case 109:  fun = precompute<109>;  break;
        case 110:  fun = precompute<110>;  break;
        case 111:  fun = precompute<111>;  break;
        case 112:  fun = precompute<112>;  break;
        case 113:  fun = precompute<113>;  break;
        case 114:  fun = precompute<114>;  break;
        case 115:  fun = precompute<115>;  break;
        case 116:  fun = precompute<116>;  break;
        case 117:  fun = precompute<117>;  break;
        case 118:  fun = precompute<118>;  break;
        case 119:  fun = precompute<119>;  break;
        case 120:  fun = precompute<120>;  break;
        case 121:  fun = precompute<121>;  break;
        case 122:  fun = precompute<122>;  break;
        case 123:  fun = precompute<123>;  break;
        case 124:  fun = precompute<124>;  break;
        case 125:  fun = precompute<125>;  break;
        case 126:  fun = precompute<126>;  break;
        case 127:  fun = precompute<127>;  break;
        case 128:  fun = precompute<128>;  break;
        case 214:  fun = precompute<214>;  break;
        case 217:  fun = precompute<217>;  break;
        case 219:  fun = precompute<219>;  break;
        case 221:  fun = precompute<221>;  break;
        case 223:  fun = precompute<223>;  break;
        case 224:  fun = precompute<224>;  break;
        case 226:  fun = precompute<226>;  break;
        case 227:  fun = precompute<227>;  break;
        case 228:  fun = precompute<228>;  break;
        case 229:  fun = precompute<229>;  break;
        case 232:  fun = precompute<232>;  break;
        case 233:  fun = precompute<233>;  break;
        case 235:  fun = precompute<235>;  break;
        case 236:  fun = precompute<236>;  break;
        case 237:  fun = precompute<237>;  break;
        case 239:  fun = precompute<239>;  break;
        case 240:  fun = precompute<240>;  break;
        case 241:  fun = precompute<241>;  break;
        case 242:  fun = precompute<242>;  break;
        case 244:  fun = precompute<244>;  break;
        case 245:  fun = precompute<245>;  break;
        case 246:  fun = precompute<246>;  break;
        case 247:  fun = precompute<247>;  break;
        case 248:  fun = precompute<248>;  break;
        case 249:  fun = precompute<249>;  break;
        case 250:  fun = precompute<250>;  break;
        case 251:  fun = precompute<251>;  break;
        case 252:  fun = precompute<252>;  break;
        case 253:  fun = precompute<253>;  break;
        case 254:  fun = precompute<254>;  break;
        case 255:  fun = precompute<255>;  break;
        case 256:  fun = precompute<256>;  break;
        case 482:  fun = precompute<482>;  break;
        case 483:  fun = precompute<483>;  break;
        case 491:  fun = precompute<491>;  break;
        case 492:  fun = precompute<492>;  break;
        case 493:  fun = precompute<493>;  break;
        case 495:  fun = precompute<495>;  break;
        case 496:  fun = precompute<496>;  break;
        case 497:  fun = precompute<497>;  break;
        case 499:  fun = precompute<499>;  break;
        case 500:  fun = precompute<500>;  break;
        case 502:  fun = precompute<502>;  break;
        case 504:  fun = precompute<504>;  break;
        case 505:  fun = precompute<505>;  break;
        case 506:  fun = precompute<506>;  break;
        case 507:  fun = precompute<507>;  break;
        case 508:  fun = precompute<508>;  break;
        case 509:  fun = precompute<509>;  break;
        case 510:  fun = precompute<510>;  break;
        case 511:  fun = precompute<511>;  break;
        case 512:  fun = precompute<512>;  break;
        case 1008: fun = precompute<1008>; break;
        case 1009: fun = precompute<1009>; break;
        case 1010: fun = precompute<1010>; break;
        case 1013: fun = precompute<1013>; break;
        case 1016: fun = precompute<1016>; break;
        case 1017: fun = precompute<1017>; break;
        case 1018: fun = precompute<1018>; break;
        case 1019: fun = precompute<1019>; break;
        case 1020: fun = precompute<1020>; break;
        case 1021: fun = precompute<1021>; break;
        case 1022: fun = precompute<1022>; break;
        case 1023: fun = precompute<1023>; break;
    }
    if (!fun) {
        std::ostringstream ss;
        ss << "template instantiation for n=" << params.n << " not available; recompilation required";
        throw std::length_error(ss.str());
    }

    fun();
}

