
#pragma once

// Note: Reusing the bitsliced implementation for just a single
// input is quite a hack, but assuming the filtering is indeed
// effective, this code will only be run every once in a while.
// Hence there's no real point in doing it properly.

#include "bitsliced.hpp"

#include <set>
#include <bitset>

template <size_t N>
using single_hullenum_t = std::array<uint8_t, N+1>;

template <size_t N>
std::pair<single_hullenum_t<N>, std::array<single_hullenum_t<N>,N>>
compute_hullenum_and_puncturedenums(std::array<std::array<bool,N>,N> const &code, std::bitset<N> const &puncture = 0)
{
    using CODE = bitsliced::code_t<N, 8, 8>;

    // allocate these on the heap to prevent stack overflows
    auto _mat = std::make_unique<typename CODE::rows_t>(); auto &mat = *_mat;
    for (size_t i = 0; i < N; ++i) {
        for (size_t j = 0; j < N; ++j)
            if (!puncture[j])
                mat[i][j] = code[i][j];
    }

    CODE::MAT::rref(mat);  //FIXME this should really be done by the caller

    auto _hulls = std::make_unique<typename CODE::hulls_t>(); auto &hulls = *_hulls;
    auto is_okay = CODE::hull_and_punctured_hulls(hulls, mat);
    assert(is_okay); (void) is_okay;

    auto _enums = std::make_unique<typename CODE::hullenums_t>(); auto &enums = *_enums;
    CODE::enumerators(enums, hulls);

    single_hullenum_t<N> hullenum;
    std::array<single_hullenum_t<N>, N> puncturedenums;

    for (size_t j = 0; j <= N; ++j) {
        for (size_t w = 0; w <= CODE::hullenum_t::count_t::max; ++w) {
            if (enums.hullenum.counts[j].eq(w))
                hullenum[j] = w;
            for (size_t i = 0; i < puncturedenums.size(); ++i)
                if (enums.punctured[i].counts[j].eq(w))
                    puncturedenums[i][j] = w;
        }
    }

    return {hullenum, puncturedenums};
}

////////////////////////////////////////////////////////////////

#include "hash.hpp"

#include <map>
#include <unordered_set>
#include <bitset>
#include <ranges>
#include <execution>

template <size_t N>
struct supportsplitting
{
    using subset_t = std::bitset<N>;
    using key_t = uint64_t;

    using tab_t = std::map<key_t, subset_t>;

    using precomp_data_t = std::pair<
            std::vector<std::set<key_t>>,
            std::vector<std::map<key_t, size_t>>
        >;

    //FIXME should check for hash collisions...
    static precomp_data_t precompute(std::array<std::array<bool,N>,N> const &code)
    {
        std::vector<tab_t> tabs(1);

        // first table: singly punctured hull enumerators
        {
            auto &tab = tabs.back();
            auto [_, puncturedenums] = compute_hullenum_and_puncturedenums<N>(code);
            assert(puncturedenums.size() == N);
            for (size_t j = 0; j < N; ++j)
                tab[myhash(puncturedenums[j])][j] = true;
        }

        std::vector<subset_t> strat;  // these are subsets of the current keys, not columns!

        bool progress = true;
        do {
            auto const &tab = tabs.back();
//std::cerr << std::setw(4) << tab.size() << "/" << N << "\r" << std::flush;

            std::vector<key_t> allkeys;
            for (auto const &[k,_]: tab)
                allkeys.push_back(k);

            assert(std::all_of(allkeys.begin(), allkeys.end(), [&](auto const &k) { return tab.at(k).count() >= 1; }));
            if (std::all_of(allkeys.begin(), allkeys.end(), [&](auto const &k) { return tab.at(k).count() == 1; }))
                break;

            progress = false;

            // given a puncturing, return the resulting refined partition
            auto const process = [&](auto const &punct) -> tab_t
            {
                auto [_, newpuncturedenums] = compute_hullenum_and_puncturedenums<N>(code, punct);
                tab_t newtab;
                for (auto const &[k,js]: tab) {
                    if (js.count() == 1) {
                        newtab[k] = js;
                        continue;
                    }
                    for (size_t j = 0; j < N; ++j)
                        if (js[j]) {
                            auto h = myhash(myhash(k), myhash(newpuncturedenums[j]));
                            newtab[h][j] = true;
                        }
                }

                return newtab;
            };

            // subset of which hullenums to puncture, resulting mapping table
            std::vector<std::pair<subset_t, tab_t>> results;

            {
                // enumerate subsets of keys of "small" cardinality
                std::vector<subset_t> subsets;
                {
                    using rec_t = std::function<void(size_t left, size_t, subset_t &)>;
                    rec_t const rec = [&](size_t left, size_t start, subset_t &keys) -> void {
                        assert(start <= allkeys.size());
                        if (start == allkeys.size()) {
                            subsets.push_back(keys);
                            return;
                        }
                        rec(left-0, start+1, keys);
                        if (!left)
                            return;
                        keys[start] = true;
                        rec(left-1, start+1, keys);
                        keys[start] = false;
                    };
                    size_t w = allkeys.size() <=   4 ? allkeys.size()
                             : allkeys.size() <=  16 ? 4
                             : allkeys.size() <=  32 ? 3
                             : allkeys.size() <=  64 ? 2
                             :                         1;
                    subset_t tmp;
                    rec(w, 0, tmp);
                }

                // for each of them, check how well the associated puncturing partitions
                results.resize(subsets.size());
                std::transform(std::execution::par_unseq, subsets.begin(), subsets.end(), results.begin(), [&](subset_t const &keys) {
                        subset_t punct {};
                        for (size_t j = 0; j < allkeys.size(); ++j)
                            if (keys[j])
                                punct |= tab.at(allkeys[j]);
                        auto newtab = process(punct);
                        if (newtab.size() > tab.size())
                            return std::make_pair(keys, newtab);
                        return std::make_pair(subset_t{}, tab_t{});
                    });

                // throw out the useless puncturings
                for (auto it = results.begin(); it != results.end(); ) {
                    if (it->first.any())
                        ++it;
                    else
                        it = results.erase(it);
                }
            }

            // among all puncturings tested above, find the best one
            if (!results.empty()) {
                auto const &[keys, newtab] = *std::min_element(results.begin(), results.end(), [](auto const &tup1, auto const &tup2) {
                        if (tup1.second.size() > tup2.second.size() ) return true;
                        if (tup1.second.size() < tup2.second.size() ) return false;
                        size_t max1 = 0, max2 = 0;
                        for (auto const &[_,v]: tup1.second) max1 = std::max(max1, v.size());
                        for (auto const &[_,v]: tup2.second) max2 = std::max(max2, v.size());
                        if (max1 < max2) return true;
                        if (max1 > max2) return false;
                        if (tup1.first.count() < tup2.first.count()) return true;
                        if (tup1.first.count() > tup2.first.count()) return false;
                        return false;
                    });
                strat.push_back(keys);
                tabs.push_back(std::move(newtab));
                progress = true;
            }

        } while (progress);
//std::cerr << "\r\x1b[K" << std::flush;

        if (!progress)
            throw std::logic_error("stuck in precomputation for support-splitting algorithm and guessing isn't implemented");

        // slightly streamline the data format for the online phase: it needs less information
        precomp_data_t out;
        auto &[outstrat, outtabs] = out;
        for (size_t i = 0; i < strat.size(); ++i) {
            auto &outkeys = outstrat.emplace_back();
            auto &outtab = outtabs.emplace_back();
            size_t j = 0;
            for (auto const &[k,js]: tabs[i]) {
                if (strat[i][j++])
                    outkeys.insert(k);
                outtab[k] = js.count();
            }
        }
        // last table is different, it contains indices instead of counts (which are all 1 at this point)
        {
            auto &outtab = outtabs.emplace_back();
            for (auto const &[k,js]: tabs.back()) {
                size_t j = 0;
                assert(js.count() == 1);
                while (!js[j]) ++j;
                outtab[k] = j;
            }
        }

        return out;
    }

    static
    std::optional<std::vector<size_t>> run(std::array<std::array<bool,N>,N> const &code, precomp_data_t const &precomp)
    {
        auto const &[strat,tabs] = precomp;
        assert(tabs.size() == strat.size() + 1);

        tab_t mytab;
        {
            auto [_, puncturedenums] = compute_hullenum_and_puncturedenums<N>(code);
            assert(puncturedenums.size() == N);
            for (size_t j = 0; j < N; ++j)
                mytab[myhash(puncturedenums[j])][j] = true;
        }

        for (size_t i = 0; i < strat.size(); ++i) {
            auto const &tab = tabs[i];

#ifndef NDEBUG
{
size_t count = 0;
for (auto [_,js]: mytab)
    count += js.count();
assert(count == N);
}
#endif

            for (auto const &[key,cnt]: tab)
                if (mytab[key].count() != cnt)
                    return {};  // not equivalent

            subset_t punct {};
            for (auto const &k: strat[i])
                punct |= mytab.at(k);

            auto [_, newpuncturedenums] = compute_hullenum_and_puncturedenums<N>(code, punct);

            tab_t newtab;
            for (auto const &[k,js]: mytab) {
                if (js.count() == 1) {
                    newtab[k] = js;
                    continue;
                }
                for (size_t j = 0; j < N; ++j)
                    if (js[j]) {
                        auto h = myhash(myhash(k), myhash(newpuncturedenums[j]));
                        newtab[h][j] = true;
                    }
            }
            mytab = std::move(newtab);
        }

        std::vector<size_t> perm(N, -1);
        {
            auto const &tab = tabs.back();

            if (!std::all_of(tab.begin(), tab.end(), [&](auto const &tup) { return mytab[tup.first].count() == 1; }))
                return {};  // not equivalent
#ifndef NDEBUG
{
size_t count = 0;
for (auto [_,js]: mytab)
    count += js.count();
assert(count == N);
}
#endif

            for (auto const &[key,i]: tab) {
                assert(perm[i] == (size_t) -1);
                size_t j = 0;
                while (!mytab.at(key)[j]) ++j;
                perm[i] = j;
            }
        }

        return perm;
    }

};  // struct supportsplitting

